#include "stdafx.h"
#include "LeoHelpers_Scope.h"
#include "LeoHelpers_String.h"
#include "LeoHelpers_File.h"
#include "LeoHelpers_Thread.h"

enum Action
{
	ACTION_UNKNOWN,
	ACTION_TEST,
	ACTION_CONNECT,
	ACTION_DISCONNECT,
};

bool ParseCommandLine(int argc, wchar_t *argv[], Action &action, std::wstring &strAdapterName, std::wstring &strOpenVpnConfig, std::wstring &strOpenVpnExe, bool &fVerbose);
bool IsNetworkAdapterConnected(const wchar_t *szAdapterName);
bool WaitForConnectionState(const wchar_t *szOpenVpnExe, const wchar_t *szAdapterName, const bool fVerbose, const bool fConnecting);
bool RunOpenVPNExe(const std::wstring &strOpenVpnExe, const wchar_t *szCommand, const std::wstring &strOpenVpnConfig, const bool fVerbose, const bool fWait);

int wmain(int argc, wchar_t *argv[])
{
	LeoHelpers::CoInitScoper coInit;
	LeoHelpers::StringInitScoper strInit;

	Action action = ACTION_UNKNOWN;
	std::wstring strAdapterName;
	std::wstring strOpenVpnConfig;
	std::wstring strOpenVpnExe;
	bool fVerbose = false;

	if (!ParseCommandLine(argc, argv, action, strAdapterName, strOpenVpnConfig, strOpenVpnExe, fVerbose))
	{
		return 1;
	}

	if (fVerbose)
	{
		wprintf(L"ConnectOpenVPN: Checking state of \"%s\" network adapter...\n", strAdapterName.c_str());
	}

	bool fIsConnected = IsNetworkAdapterConnected(strAdapterName.c_str());

	if (fVerbose)
	{
		wprintf(L"ConnectOpenVPN: \"%s\" network adapter is %sconnected.\n", strAdapterName.c_str(), fIsConnected?L"":L"not ");
	}

	if (action == ACTION_TEST)
	{
		if (fVerbose)
		{
			wprintf(L"ConnectOpenVPN: TEST action finshed.\n");
		}

		return fIsConnected ? 0 : 1;
	}

	if (action == ACTION_CONNECT)
	{
		if (fIsConnected)
		{
			if (fVerbose)
			{
				wprintf(L"ConnectOpenVPN: Already connected, so nothing to do.\n");
			}
			return 0; // Success. Nothing needs doing.
		}

		// Don't wait for it to complete, so we have a better chance of seeing the UI appear.
		if (!RunOpenVPNExe(strOpenVpnExe, L"connect", strOpenVpnConfig, fVerbose, false))
		{
			return 1;
		}

		if (!WaitForConnectionState(strOpenVpnExe.c_str(), strAdapterName.c_str(), fVerbose, true))
		{
			return 1;
		}

		if (fVerbose)
		{
			wprintf(L"ConnectOpenVPN: CONNECT action finshed.\n");
		}

		return 0;
	}

	if (action == ACTION_DISCONNECT)
	{
		if (!fIsConnected)
		{
			if (fVerbose)
			{
				wprintf(L"ConnectOpenVPN: Already disconnected, so nothing to do.\n");
			}
			return 0; // Success. Nothing needs doing.
		}

		// Wait for it to complete as we don't look for a UI here.
		if (!RunOpenVPNExe(strOpenVpnExe, L"disconnect", strOpenVpnConfig, fVerbose, true))
		{
			return 1;
		}

		if (!WaitForConnectionState(strOpenVpnExe.c_str(), strAdapterName.c_str(), fVerbose, false))
		{
			return 1;
		}

		if (fVerbose)
		{
			wprintf(L"ConnectOpenVPN: DISCONNECT action finshed.\n");
		}

		return 0;
	}

	return 1;
}

bool IsString(const wchar_t *szStringToTest, const wchar_t *szString1, const wchar_t *szString2 = nullptr, const wchar_t *szString3 = nullptr)
{
	if (!szStringToTest || !szString1)
	{
		return false;
	}

	if (szString1 && _wcsicmp(szStringToTest, szString1) == 0)
	{
		return true;
	}

	if (szString2 && _wcsicmp(szStringToTest, szString2) == 0)
	{
		return true;
	}

	if (szString3 && _wcsicmp(szStringToTest, szString3) == 0)
	{
		return true;
	}

	return false;
}

bool IsHelpString(const wchar_t *sz)
{
	while(*sz == L'-' || *sz == L'/')
	{
		++sz;
	}

	return IsString(sz, L"h", L"help", L"?");
}

bool ParseCommandLine(int argc, wchar_t *argv[], Action &action, std::wstring &strAdapterName, std::wstring &strOpenVpnConfig, std::wstring &strOpenVpnExe, bool &fVerbose)
{
	const wchar_t *szDefaultExe = L"C:\\Program Files\\OpenVPN\\bin\\openvpn-gui.exe";

	if (argc > 0)
	{
		// Skip first argument (our exe path).
		--argc;
		++argv;
	}

	if (argc == 0
	||	(argc > 0 && IsHelpString(argv[0])))
	{
		wprintf(L"ConnectOpenVPN usage:\n");
		wprintf(L" Actions (exactly one must be specified):\n");
		wprintf(L"   /test       : Test if network adapter connected. Sets success code and exits.\n");
		wprintf(L"   /connect    : If network adapter not already connected, tell OpenVPN to connect and wait for connection window to close.\n");
		wprintf(L"   /disconnect : If network adapter connected, tell OpenVPN to disconnect.\n");
		wprintf(L" Parameters:\n");
		wprintf(L"   /adapter x  : [Required] Name of OpenVPN's network adapter, from the Windows Network Adapters control panel.\n");
		wprintf(L"   /config x   : [Required except for /test] Name (not path) of OpenVPN config file; .ovpn extension optional.\n");
		wprintf(L"   /guiexe x   : [Optional] Full path to openvpn-gui.exe, if it is not in the normal location.\n");
		wprintf(L"   /verbose    : Print diagnostic information for debugging. Normally, there is no output unless there is an error.\n");
		wprintf(L" Example:\n");
		wprintf(L"  ConnectOpenVPN.exe /connect /adapter \"OpenVPN\" /config \"MyVPN.ovpn\"\n");

		return false;
	}

	action = ACTION_UNKNOWN;
	strAdapterName.clear();
	strOpenVpnConfig.clear();
	strOpenVpnExe.clear();
	fVerbose = false;

	for(int i = 0; i < argc; ++i)
	{
		const wchar_t *szSwitch = argv[i];

		if (IsString(szSwitch, L"/test"))
		{
			if (action != ACTION_UNKNOWN)
			{
				wprintf(L"ConnectOpenVPN failed: Multiple actions specified on the same command line.\n");
				return false;
			}
			action = ACTION_TEST;
		}
		else
		if (IsString(szSwitch, L"/connect"))
		{
			if (action != ACTION_UNKNOWN)
			{
				wprintf(L"ConnectOpenVPN failed: Multiple actions specified on the same command line.\n");
				return false;
			}
			action = ACTION_CONNECT;
		}
		else
		if (IsString(szSwitch, L"/disconnect"))
		{
			if (action != ACTION_UNKNOWN)
			{
				wprintf(L"ConnectOpenVPN failed: Multiple actions specified on the same command line.\n");
				return false;
			}
			action = ACTION_DISCONNECT;
		}
		else
		if (IsString(szSwitch, L"/adapter", L"/adaptor"))
		{
			if (!strAdapterName.empty())
			{
				wprintf(L"ConnectOpenVPN failed: Multiple network adapters supplied on the same command line.\n");
				return false;
			}

			if (i+1 >= argc)
			{
				wprintf(L"ConnectOpenVPN failed: The /adapter switch must be followed by a network adapter name.\n");
				return false;
			}

			strAdapterName = argv[++i];
		}
		else
		if (IsString(szSwitch, L"/config"))
		{
			if (!strOpenVpnConfig.empty())
			{
				wprintf(L"ConnectOpenVPN failed: Multiple OpenVPN config names supplied on the same command line.\n");
				return false;
			}

			if (i+1 >= argc)
			{
				wprintf(L"ConnectOpenVPN failed: The /config switch must be followed by an OpenVPN config name.\n");
				return false;
			}

			strOpenVpnConfig = argv[++i];
		}
		else
		if (IsString(szSwitch, L"/guiexe"))
		{
			if (!strOpenVpnExe.empty())
			{
				wprintf(L"ConnectOpenVPN failed: Multiple openvpn-gui.exe paths supplied on the same command line.\n");
				return false;
			}

			if (i+1 >= argc)
			{
				wprintf(L"ConnectOpenVPN failed: The /guiexe switch, if used at all, must be followed by the full path to openvpn-gui.exe.\n");
				return false;
			}

			strOpenVpnExe = argv[++i];
		}
		else
		if (IsString(szSwitch, L"/verbose"))
		{
			fVerbose = true;
		}
		else
		{
			wprintf(L"ConnectOpenVPN failed: Unrecognised argument \"%s\".\n", szSwitch);
			wprintf(L" (Remember to surround paths and other arguments containing spaces \"in quotes\".)\n");
			return false;
		}
	}

	if (action == ACTION_UNKNOWN)
	{
		wprintf(L"ConnectOpenVPN failed: No action specified.\n");
		return false;
	}

	if (strAdapterName.empty())
	{
		wprintf(L"ConnectOpenVPN failed: No network adapter specified.\n");
		return false;
	}

	if (strAdapterName.empty())
	{
		wprintf(L"ConnectOpenVPN failed: No network adapter specified.\n");
		return false;
	}

	if (action != ACTION_TEST && strOpenVpnConfig.empty())
	{
		wprintf(L"ConnectOpenVPN failed: No OpenVPN config name specified.\n");
		return false;
	}

	if (strOpenVpnExe.empty())
	{
		strOpenVpnExe = szDefaultExe;
	}

	if (action != ACTION_TEST && LeoHelpers::GetFileAttributes(strOpenVpnExe.c_str()) == INVALID_FILE_ATTRIBUTES)
	{
		wprintf(L"ConnectOpenVPN failed: Cannot find \"%s\".\n", strOpenVpnExe.c_str());
		if (wcscmp(strOpenVpnExe.c_str(), szDefaultExe) == 0)
		{
			wprintf(L" (You can use /guiexe to specify the path if it is not in the normal location.)\n");
		}
		return false;
	}

	return true;
}

// Returns ERROR_SUCCESS on success.
// If callers use fields that differ or are missing between IP_ADAPTER_ADDRESSES_XP and IP_ADAPTER_ADDRESSES_LH
// (or between the different versions of IP_ADAPTER_ADDRESSES_LH), remember to test the Length field as appropriate.
ULONG GetAdaptersAddressesWrapper(ULONG ulFamily, ULONG ulFlags, LeoHelpers::AutoBuffer< IP_ADAPTER_ADDRESSES > &adrs)
{
	// Ensure adrs is freed consistently on all failure paths.
	LeoHelpers::FreeAutoBufferOnDestruction autoFree{adrs};

	for(int iRetry = 0; true; ++iRetry)
	{
		ULONG ulError = ERROR_SUCCESS;
		ULONG ulBufferBytes = 0;

		// Query the required buffer size. Expect this to return ERROR_BUFFER_OVERFLOW, or ERROR_NO_DATA.
		ulError = GetAdaptersAddresses(ulFamily, ulFlags, NULL, NULL, &ulBufferBytes);

		if (ulError != ERROR_BUFFER_OVERFLOW || ulBufferBytes == 0)
		{
			if (ulError == ERROR_SUCCESS)
			{
				// Paranoia. API should already return ERROR_NO_DATA when no matching adapters, but just in case.
				ulError = ERROR_NO_DATA;
			}

			return ulError;
		}

		// Allocate buffer. If we already have a larger buffer then keep it.
		if (!adrs.AllocateMinimumBytes(ulBufferBytes))
		{
			return ERROR_NOT_ENOUGH_MEMORY; // Consistent with what docs say GetAdaptersAddresses returns when out of memory.
		}
		ulBufferBytes = static_cast< ULONG >( min(adrs.GetSizeBytesInUse(), ULONG_MAX) );

		// Get the data into the buffer.
		ulError = GetAdaptersAddresses(ulFamily, ulFlags, NULL, adrs.GetBuffer(), &ulBufferBytes);

		// If a network adapter was added between the two GetAdaptersAddresses calls, then we might need to loop for a bigger buffer.
		if (ulError == ERROR_BUFFER_OVERFLOW)
		{
			// Don't keep trying forever. It's unlikely something would keep adding new adapters faster than we can query the API.
			if (iRetry == 9)
			{
				return ulError;
			}

			continue;
		}

		if (ulError == ERROR_SUCCESS)
		{
			// Prevent adrs from being freed.
			autoFree.Forget();
		}

		return ulError;
	}
}

bool IsNetworkAdapterConnected(const wchar_t *szAdapterName)
{
	if (!szAdapterName || !*szAdapterName)
	{
		return false;
	}

	LeoHelpers::AutoBuffer< IP_ADAPTER_ADDRESSES > adrs;
	const ULONG ulFlags = GAA_FLAG_SKIP_UNICAST|GAA_FLAG_SKIP_ANYCAST|GAA_FLAG_SKIP_MULTICAST|GAA_FLAG_SKIP_DNS_SERVER;
	if (GetAdaptersAddressesWrapper(AF_UNSPEC, ulFlags, adrs) != ERROR_SUCCESS)
	{
		return false;
	}

	// We only use pAdr->Next, FriendlyName and OperStatus, which are in the same places in all versions of the
	// IP_ADAPTER_ADDRESSES structures, so there is no need to test pAdr->Length.

	for(const IP_ADAPTER_ADDRESSES *pAdr = adrs.GetBuffer(); pAdr; pAdr = pAdr->Next)
	{
		if (_wcsicmp(szAdapterName, pAdr->FriendlyName) == 0)
		{
			return pAdr->OperStatus == IfOperStatusUp;
		}
	}

	return false;
}

class FindProcessWindows
{
	const std::wstring m_strProcessPath;
	const std::wstring m_strClassName;
	const bool m_fVisibleOnly;

	std::map< int, std::wstring > m_mapPIDtoPath;
	std::set< HWND > m_setFound;

	// This buffer is used when calling QueryFullProcessImageName. Since we only want to know if it's equal to
	// the requested path, a buffer that is 2 characters longer (one for the null terminator, and one for any
	// extra character beyond the length of our desired path) is sufficient: Either the API call will fail
	// or the API will truncate the longer path and it still won't be identical to the desired path.
	LeoHelpers::WCharArrayScoper m_scPathBuffer;
	size_t PathBufferLength() const
	{
		return m_strProcessPath.length() + 2;
	}

	static BOOL CALLBACK EnumProc(HWND hwnd, LPARAM lParam)
	{
		FindProcessWindows *pThis = reinterpret_cast<FindProcessWindows *>(lParam);

		if (pThis->m_fVisibleOnly && !IsWindowVisible(hwnd))
		{
			return TRUE;
		}

		if (!pThis->m_strClassName.empty())
		{
			wchar_t szClassName[257] = {L'\0'};
			if (!GetClassName(hwnd, szClassName, _countof(szClassName)-1))
			{
				return TRUE;
			}

			if (_wcsicmp(szClassName, pThis->m_strClassName.c_str()) != 0)
			{
				return TRUE;
			}
		}

		if (!pThis->m_strProcessPath.empty() && !pThis->m_scPathBuffer.IsEmpty())
		{
			DWORD dwPID = 0;
			if (!GetWindowThreadProcessId(hwnd, &dwPID) || !dwPID)
			{
				return TRUE;
			}

			const auto miter = pThis->m_mapPIDtoPath.find(dwPID);

			if (miter != pThis->m_mapPIDtoPath.end())
			{
				// miter->second may be empty here, in which case we cached the failure and it won't succeed the comparison test.

				if (_wcsicmp(miter->second.c_str(), pThis->m_strProcessPath.c_str()) != 0)
				{
					return TRUE;
				}
			}
			else
			{
				pThis->m_scPathBuffer.Get()[0] = L'\0';

				LeoHelpers::HandleScoper scProcHandle{ OpenProcess(PROCESS_QUERY_INFORMATION, FALSE, dwPID) };
				if (scProcHandle.HasHandle())
				{
					DWORD dwBufferSize = static_cast<DWORD>(pThis->PathBufferLength()); // It's checked against MAXDWORD in the constructor.
					if (!QueryFullProcessImageName(scProcHandle.Get(), 0, pThis->m_scPathBuffer.Get(), &dwBufferSize))
					{
						pThis->m_scPathBuffer.Get()[0] = L'\0';
					}
					scProcHandle.Close();
				}

				// pThis->m_scPathBuffer.Get() may be empty here, in which case we cache the failure and it won't succeed the comparison test.

				pThis->m_mapPIDtoPath[dwPID] = pThis->m_scPathBuffer.Get();

				if (_wcsicmp(pThis->m_scPathBuffer.Get(), pThis->m_strProcessPath.c_str()) != 0)
				{
					return TRUE;
				}
			}
		}

		pThis->m_setFound.insert(hwnd);

		return TRUE;
	}

public:
	FindProcessWindows(const wchar_t *szProcessPath, const wchar_t *szClassName, bool fVisibleOnly)
		: m_strProcessPath(szProcessPath?szProcessPath:L"")
		, m_strClassName(szClassName?szClassName:L"")
		, m_fVisibleOnly(fVisibleOnly)
	{
		if (!m_strProcessPath.empty() && PathBufferLength() < MAXDWORD)
		{
			m_scPathBuffer.Set( new wchar_t[ PathBufferLength() ] );
		}
	}

	bool Enum()
	{
		m_mapPIDtoPath.clear();
		m_setFound.clear();

		BOOL fRes = EnumWindows(EnumProc, reinterpret_cast<LPARAM>(this));
		m_mapPIDtoPath.clear();

		return !!fRes;
	}

	bool IsEmpty()
	{
		return m_setFound.empty();
	}

	void TakeResults(std::set< HWND > & setFound)
	{
		setFound = std::move(m_setFound);
		m_setFound.clear();
	}
};

bool WaitForConnectionState(const wchar_t *szOpenVpnExe, const wchar_t *szAdapterName, const bool fVerbose, const bool fConnecting)
{
	if (!szAdapterName || !*szAdapterName)
	{
		return false;
	}

	const int iSleepTimeFast = 50;
	const int iSleepTime = 250;

	if (fConnecting)
	{
		if (fVerbose)
		{
			wprintf(L"ConnectOpenVPN: Waiting for OpenVPN GUI to appear...\n");
		}

		FindProcessWindows finder{szOpenVpnExe, L"#32770", true};

		// Wait ~10 seconds for the UI to open.
		for(int iRetry = 0; iRetry < 10000/iSleepTimeFast; ++iRetry)
		{
			if (!finder.Enum())
			{
				wprintf(L"ConnectOpenVPN failed: Error enumerating windows.\n");
				return false;
			}
				
			if (!finder.IsEmpty())
			{
				if (fVerbose)
				{
					wprintf(L"ConnectOpenVPN: OpenVPN GUI found.\n");
				}
				break;
			}

			Sleep(iSleepTimeFast);
		}

		if (finder.IsEmpty())
		{
			if (IsNetworkAdapterConnected(szAdapterName))
			{
				wprintf(L"ConnectOpenVPN warning: OpenVPN GUI was not detected, but network is now connected. May require a code-change.\n");
				return true;
			}

			wprintf(L"ConnectOpenVPN failed: OpenVPN GUI did not appear or took too long.\n");
			return false;
		}

		// Wait an infinite amount of time for the UI to close. Allows for password prompts.
		// We wait for the Ui to close before testing the connection status, as that is the point the network is really usable.
		// The adapter will go into a "connected" state a short time before it is actually ready to start processing data
		// and before the UI closes.

		if (fVerbose)
		{
			wprintf(L"ConnectOpenVPN: Waiting for OpenVPN GUI to close...\n");
		}

		for(int iRetry = 0; true; ++iRetry)
		{
			if (!finder.Enum())
			{
				wprintf(L"ConnectOpenVPN failed: Error enumerating windows.\n");
				return false;
			}
				
			if (finder.IsEmpty())
			{
				if (fVerbose)
				{
					wprintf(L"ConnectOpenVPN: OpenVPN GUI closed.\n");
				}
				break;
			}

			Sleep(iSleepTime);
		}

		if (fVerbose)
		{
			wprintf(L"ConnectOpenVPN: Waiting for/confirming VPN connection exists...\n");
		}

		// Wait ~2 seconds for the connection.
		for(int iRetry = 0; iRetry < 2000/iSleepTime; ++iRetry)
		{
			if (IsNetworkAdapterConnected(szAdapterName))
			{
				if (fVerbose)
				{
					wprintf(L"ConnectOpenVPN: Connected.\n");
				}
				return true;
			}

			Sleep(iSleepTime);
		}

		wprintf(L"ConnectOpenVPN failed: Connection failed or took too long.\n");
		return false;
	}
	else
	{
		if (fVerbose)
		{
			wprintf(L"ConnectOpenVPN: Waiting for disconnection...\n");
		}

		// Wait ~5 seconds for the disconnection.
		for(int iRetry = 0; iRetry < 5000/iSleepTime; ++iRetry)
		{
			if (!IsNetworkAdapterConnected(szAdapterName))
			{
				if (fVerbose)
				{
					wprintf(L"ConnectOpenVPN: Disconnected.\n");
				}
				return true;
			}

			if (fVerbose && iRetry == 0)
			{
				wprintf(L"ConnectOpenVPN: Waiting for disconnection...\n");
			}

			Sleep(iSleepTime);
		}

		wprintf(L"ConnectOpenVPN failed: Disconnection failed or took too long.\n");
		return false;
	}
}

bool RunOpenVPNExe(const std::wstring &strOpenVpnExe, const wchar_t *szCommand, const std::wstring &strOpenVpnConfig, const bool fVerbose, const bool fWait)
{
	std::wstring strArgs = L"--command ";
	strArgs += szCommand;
	strArgs += L" \"";
	strArgs += strOpenVpnConfig;
	strArgs += L"\"";

	if (fVerbose)
	{
		wprintf(L"ConnectOpenVPN: Running: \"%s\" %s\n", strOpenVpnExe.c_str(), strArgs.c_str());
	}

	LeoHelpers::HandleScoper scProcess;
	if (!LeoHelpers::CreateProcessWrapper(strOpenVpnExe.c_str(), strArgs.c_str(), false, nullptr, true, scProcess.GetAssignable(), nullptr, false))
	{
		wprintf(L"ConnectOpenVPN failed: Error running \"%s\" %s\n", strOpenVpnExe.c_str(), strArgs.c_str());
		return false;
	}

	if (fWait)
	{
		if (fVerbose)
		{
			wprintf(L"ConnectOpenVPN: Waiting for OpenVPN exe to return.\n");
		}

		if (::WaitForSingleObject(scProcess.Get(), INFINITE) != WAIT_OBJECT_0)
		{
			wprintf(L"ConnectOpenVPN failed: Error waiting for OpenVPN exe to return.\n");
			return false;
		}
	}

	return true;
}
