#pragma once

#define LEOHELPERS_SCOPE_INCLUDED

namespace LeoHelpers
{
	class InterlockedCountScoper
	{
	private:
		volatile LONG *m_pl;
	private:
		InterlockedCountScoper(const InterlockedCountScoper &rhs); // disallow
		InterlockedCountScoper &operator=(const InterlockedCountScoper &rhs); // disallow
	public:
		InterlockedCountScoper(volatile LONG *pl) : m_pl(pl) { ::InterlockedIncrement(m_pl); }
		/* not virtual */ ~InterlockedCountScoper()          { ::InterlockedDecrement(m_pl); }
	};

	class CritSecWrap
	{
	private:
		CRITICAL_SECTION m_cs;
	private:
		CritSecWrap(const CritSecWrap &rhs); // disallow
		CritSecWrap &operator=(const CritSecWrap &rhs); // disallow
	public:
		CritSecWrap()                    { ::InitializeCriticalSection(&m_cs); }
		/* not virtual */ ~CritSecWrap() { ::DeleteCriticalSection(&m_cs);     }
	//	operator CRITICAL_SECTION * ()   { return &m_cs; }    <-- Not defining this because it'd invite mistaking the wrapper for normal CS and init/deleting it.
		CRITICAL_SECTION *GetCS()        { return &m_cs; } // <-- Instead, things like CriticalSectionScoper will understand how to use this as well.
	};

	class CriticalSectionScoper
	{
	private:
		CRITICAL_SECTION *m_pCS;
	private:
		CriticalSectionScoper(const CriticalSectionScoper &rhs); // disallow
		CriticalSectionScoper &operator=(const CriticalSectionScoper &rhs); // disallow
	public:
		CriticalSectionScoper(CRITICAL_SECTION *pCS) : m_pCS(pCS)         { ::EnterCriticalSection(m_pCS); }
		CriticalSectionScoper(CritSecWrap &csw)      : m_pCS(csw.GetCS()) { ::EnterCriticalSection(m_pCS); }
		/*not virtual*/ ~CriticalSectionScoper()                          { ::LeaveCriticalSection(m_pCS); }
	};

	class CriticalSection_Try_Scoper
	{
	private:
		CRITICAL_SECTION *m_pCS;
		bool m_bSuccess;
	private:
		CriticalSection_Try_Scoper(const CriticalSection_Try_Scoper &rhs); // disallow
		CriticalSection_Try_Scoper &operator=(const CriticalSection_Try_Scoper &rhs); // disallow
	public:
		CriticalSection_Try_Scoper(CRITICAL_SECTION *pCS) : m_pCS(pCS),         m_bSuccess(false) { if (TryEnterCriticalSection(m_pCS)) { m_bSuccess = true; } }
		CriticalSection_Try_Scoper(CritSecWrap &csw)      : m_pCS(csw.GetCS()), m_bSuccess(false) { if (TryEnterCriticalSection(m_pCS)) { m_bSuccess = true; } }
		/*not virtual*/ ~CriticalSection_Try_Scoper() { if (m_bSuccess) { LeaveCriticalSection(m_pCS); } }
		bool Retry()             { if (!m_bSuccess && TryEnterCriticalSection(m_pCS)) { m_bSuccess = true; } return m_bSuccess; }
		void EnterIfNotAlready() { if (!m_bSuccess) { EnterCriticalSection(m_pCS); m_bSuccess = true; } }
		bool Succeeded() const   { return  m_bSuccess; }
		bool Failed()    const   { return !m_bSuccess; }
	};

	// This does the opposite of CriticalSectionScoper; i.e. leaves on construction and re-enters on destruction.
	// Obviously, you should only construct this object when you have entered the CS exactly once.
	class CriticalSectionScoperReverse
	{
	private:
		CRITICAL_SECTION *m_pCS;
	private:
		CriticalSectionScoperReverse(const CriticalSectionScoperReverse &rhs); // disallow
		CriticalSectionScoperReverse &operator=(const CriticalSectionScoperReverse &rhs); // disallow
	public:
		CriticalSectionScoperReverse(CRITICAL_SECTION *pCS) : m_pCS(pCS)    { LeaveCriticalSection(m_pCS); }
		CriticalSectionScoperReverse(CritSecWrap &csw) : m_pCS(csw.GetCS()) { LeaveCriticalSection(m_pCS); }
		/*not virtual*/ ~CriticalSectionScoperReverse()                     { EnterCriticalSection(m_pCS); }
	};

	template< typename T > class CriticalObject
	{
	private:
		mutable CRITICAL_SECTION m_cs;
	private:
		CriticalObject(const CriticalObject< T > &rhs); // disallow
		CriticalObject &operator=(const CriticalObject< T > &rhs); // disallow
	public:
		CriticalObject(T t) : object(t)   { InitializeCriticalSection(&m_cs); }
		/*not virtual*/ ~CriticalObject() { DeleteCriticalSection(&m_cs); }
		void Lock() const                 { EnterCriticalSection(&m_cs); }
		void Unlock() const               { LeaveCriticalSection(&m_cs); }
		T object;
	};

	template< typename T > class CriticalObjectScoper
	{
	private:
		const CriticalObject< T > *m_pCO;
	private:
		CriticalObjectScoper(const CriticalObjectScoper< T > &rhs); // disallow
		CriticalObjectScoper &operator=(const CriticalObjectScoper< T > &rhs); // disallow
	public:
		CriticalObjectScoper(const CriticalObject< T > *pCO) : m_pCO(pCO) { m_pCO->Lock(); }
		/*not virtual*/ ~CriticalObjectScoper()                           { m_pCO->Unlock(); }
	};

	template< typename T > class GenericLockScoper
	{
	private:
		T *m_pCO;
	private:
		GenericLockScoper(const GenericLockScoper< T > &rhs); // disallow
		GenericLockScoper &operator=(const GenericLockScoper< T > &rhs); // disallow
	public:
		GenericLockScoper(T *pCO) : m_pCO(pCO) { m_pCO->Lock(); }
		/*not virtual*/ ~GenericLockScoper()         { m_pCO->Unlock(); }
	};

	class MutexWrapper
	{
	private:
		HANDLE m_hMutex;
	private:
		MutexWrapper(const MutexWrapper &rhs); // disallow
		MutexWrapper &operator=(const MutexWrapper &rhs); // disallow
	public:
		MutexWrapper()                      : m_hMutex(::CreateMutex(NULL, FALSE, NULL  )) { assert(m_hMutex != NULL); }
		MutexWrapper(const wchar_t *szName) : m_hMutex(::CreateMutexW(NULL, FALSE, szName)) { assert(m_hMutex != NULL); }
		/* not virtual */ ~MutexWrapper() { if (m_hMutex != NULL) { ::CloseHandle(m_hMutex); m_hMutex = NULL; } }
	//	operator HANDLE () { return m_hMutex; }    <-- Not defining this because it'd invite mistaking the wrapper for normal mutex and init/deleting it.
		HANDLE GetMX()     { return m_hMutex; } // <-- Instead, things like MutexScoper will understand how to use this as well.

	};

	// See also: SingleInstance class in LeoHelpers_Thread.h/cpp
	class MutexScoper
	{
	private:
		HANDLE m_hMutex;
	private:
		MutexScoper(const MutexScoper &rhs); // disallow
		MutexScoper &operator=(const MutexScoper &rhs); // disallow

		void Obtain(HANDLE hMutex, DWORD dwTimeout, HANDLE hAbortEvent)
		{
			assert(m_hMutex == NULL);
			if (hMutex != NULL)
			{
				HANDLE waitEvents[2] = { hMutex, hAbortEvent };

				DWORD dwWaitMutexRes = ::WaitForMultipleObjects(hAbortEvent ? 2 : 1, waitEvents, FALSE, dwTimeout);
				assert(dwWaitMutexRes == WAIT_OBJECT_0 || dwWaitMutexRes == (WAIT_OBJECT_0+1) || (dwTimeout != INFINITE && dwWaitMutexRes == WAIT_TIMEOUT));
				if (dwWaitMutexRes == WAIT_OBJECT_0)
				{
					m_hMutex = hMutex;
				}
			}
		}
	public:
		MutexScoper(HANDLE hMutex)                                          : m_hMutex(NULL) { Obtain(hMutex,      INFINITE,  NULL);        }
		MutexScoper(HANDLE hMutex, DWORD dwTimeout)                         : m_hMutex(NULL) { Obtain(hMutex,      dwTimeout, NULL);        }
		MutexScoper(HANDLE hMutex, DWORD dwTimeout, HANDLE hAbortEvent)     : m_hMutex(NULL) { Obtain(hMutex,      dwTimeout, hAbortEvent); }
		MutexScoper(MutexWrapper &mxw)                                      : m_hMutex(NULL) { Obtain(mxw.GetMX(), INFINITE,  NULL);        }
		MutexScoper(MutexWrapper &mxw, DWORD dwTimeout)                     : m_hMutex(NULL) { Obtain(mxw.GetMX(), dwTimeout, NULL);        }
		MutexScoper(MutexWrapper &mxw, DWORD dwTimeout, HANDLE hAbortEvent) : m_hMutex(NULL) { Obtain(mxw.GetMX(), dwTimeout, hAbortEvent); }
		/*not virtual*/ ~MutexScoper()                                                       { ReleaseNow(); }
		void ReleaseNow()                                                                    { if (m_hMutex != NULL) { BOOL bRelMutexRes = ::ReleaseMutex(m_hMutex); m_hMutex = NULL; assert(bRelMutexRes); } }
		void ReleaseAndObtain(HANDLE hMutex)                                                 { ReleaseNow(); Obtain(hMutex,      INFINITE,  NULL);        }
		void ReleaseAndObtain(HANDLE hMutex, DWORD dwTimeout)                                { ReleaseNow(); Obtain(hMutex,      dwTimeout, NULL);        }
		void ReleaseAndObtain(HANDLE hMutex, DWORD dwTimeout, HANDLE hAbortEvent)            { ReleaseNow(); Obtain(hMutex,      dwTimeout, hAbortEvent); }
		void ReleaseAndObtain(MutexWrapper &mxw)                                             { ReleaseNow(); Obtain(mxw.GetMX(), INFINITE,  NULL);        }
		void ReleaseAndObtain(MutexWrapper &mxw, DWORD dwTimeout)                            { ReleaseNow(); Obtain(mxw.GetMX(), dwTimeout, NULL);        }
		void ReleaseAndObtain(MutexWrapper &mxw, DWORD dwTimeout, HANDLE hAbortEvent)        { ReleaseNow(); Obtain(mxw.GetMX(), dwTimeout, hAbortEvent); }
		bool IsGood()  const                                                                 { return (m_hMutex != NULL); }
		bool IsEmpty() const                                                                 { return (m_hMutex == NULL); }
	};

	class SemaphoreScoper
	{
	private:
		HANDLE m_hSem;
	private:
		SemaphoreScoper(const SemaphoreScoper &rhs); // disallow
		SemaphoreScoper &operator=(const SemaphoreScoper &rhs); // disallow
	public:
		SemaphoreScoper() : m_hSem(NULL) {  }
		SemaphoreScoper(HANDLE hSem) : m_hSem(NULL) { Obtain(hSem); }
		void Obtain(HANDLE hSem) { assert(hSem != NULL && m_hSem == NULL); Release(); DWORD dwRes = ::WaitForSingleObject(hSem, INFINITE); assert(dwRes == WAIT_OBJECT_0); if (dwRes == WAIT_OBJECT_0) { m_hSem = hSem; } }
		void Release() { if (m_hSem != NULL) { BOOL bRelSemRes = ::ReleaseSemaphore(m_hSem, 1, NULL); assert(bRelSemRes); m_hSem = NULL; } }
		/*not virtual*/ ~SemaphoreScoper()          { Release(); }
	};

	class SetEventOnDestruction
	{
	private:
		HANDLE m_hEvent;
	private:
		SetEventOnDestruction(const SetEventOnDestruction &rhs); //disallow
		SetEventOnDestruction &operator=(const SetEventOnDestruction &rhs); // disallow
	public:
		SetEventOnDestruction(HANDLE hEvent) : m_hEvent(hEvent) { }
		/* not virtual */ ~SetEventOnDestruction()              { if (m_hEvent != NULL) { ::SetEvent(m_hEvent); } }
		void SetNowAndForget()                                  { if (m_hEvent != NULL) { ::SetEvent(m_hEvent); m_hEvent = NULL; } }
		void Forget()                                           { m_hEvent = NULL; }
	};

	class HandleScoper
	{
	private:
		HANDLE m_h;
	private:
		HandleScoper(const HandleScoper &rhs); // disallow
		HandleScoper &operator=(const HandleScoper &rhs); // disallow
	public:
		HandleScoper()         : m_h(INVALID_HANDLE_VALUE) { }
		HandleScoper(HANDLE h) : m_h(h)                    { }
		/* not virtual */ ~HandleScoper() { Close(); }
		inline bool   HasHandle() const   { return (m_h != NULL && m_h != INVALID_HANDLE_VALUE); }
		inline void   Forget()            { m_h = INVALID_HANDLE_VALUE; }
		inline void   Close()             { if (HasHandle()) { ::CloseHandle(m_h); } Forget(); }
		inline HANDLE Get() const         { assert(HasHandle()); return m_h; }
		inline HANDLE GetAndForget()      { assert(HasHandle()); HANDLE h = m_h; Forget(); return h; }
		inline HANDLE *GetAssignable()    { assert(!HasHandle()); Close(); return &m_h; }
		inline void   Set(HANDLE h)       { if (h != m_h) { Close(); m_h = h; } }
	};

	class HandleScoperPreserveErrors // Special version that preserves last-error during destruction.
	{
	private:
		HANDLE m_h;
	private:
		HandleScoperPreserveErrors(const HandleScoperPreserveErrors &rhs) = delete;
		HandleScoperPreserveErrors &operator=(const HandleScoperPreserveErrors &rhs) = delete;
	public:
		HandleScoperPreserveErrors()         : m_h(INVALID_HANDLE_VALUE) { }
		HandleScoperPreserveErrors(HANDLE h) : m_h(h)                    { }
		/* not virtual */ ~HandleScoperPreserveErrors()
		{
			DWORD dwErr = ::GetLastError();
			Close();
			::SetLastError(dwErr);
		}
		inline bool   HasHandle() const   { return (m_h != NULL && m_h != INVALID_HANDLE_VALUE); }
		inline void   Forget()            { m_h = INVALID_HANDLE_VALUE; }
		inline void   Close()             { if (HasHandle()) { ::CloseHandle(m_h); } Forget(); }
		inline HANDLE Get() const         { assert(HasHandle()); return m_h; }
		inline HANDLE GetAndForget()      { assert(HasHandle()); HANDLE h = m_h; Forget(); return h; }
		inline void   Set(HANDLE h)       { if (h != m_h) { Close(); m_h = h; } }
	};

	class CancelIOScoper
	{
	private:
		HANDLE m_h;
		bool m_bNeedCancel;
	private:
		CancelIOScoper(const CancelIOScoper &rhs) = delete;
		CancelIOScoper &operator=(const CancelIOScoper &rhs) = delete;
	public:
		CancelIOScoper() : m_h(INVALID_HANDLE_VALUE), m_bNeedCancel(false) { }
		CancelIOScoper(HANDLE h, bool bNeedCancel) : m_h(h), m_bNeedCancel(bNeedCancel) { }
		/* not virtual */ ~CancelIOScoper()
		{
			DWORD dwErr = ::GetLastError();
			CancelIfNeeded();
			::SetLastError(dwErr);
		}
		inline bool   HasHandle() const               { return (m_h != NULL && m_h != INVALID_HANDLE_VALUE); }
		inline bool   NeedCancel() const              { return (HasHandle() && m_bNeedCancel); }
		inline void   Forget()                        { m_h = INVALID_HANDLE_VALUE; m_bNeedCancel = false; }
		inline void   CancelIfNeeded()                { if (NeedCancel()) { ::CancelIo(m_h); } m_bNeedCancel = false; }
		inline void   SetNeedCancel(bool bNeedCancel) { m_bNeedCancel = bNeedCancel; }
	};

	template < class FileOpFuncs >
	class CloseFileScoper
	{
	private:
		FileOpFuncs &m_funcs;
		typename FileOpFuncs::FileHandleType m_handle;
	private:
		CloseFileScoper(const CloseFileScoper &rhs); // disallow
		CloseFileScoper &operator=(const CloseFileScoper &rhs); // disallow
	public:
		CloseFileScoper(FileOpFuncs &funcs) : m_funcs(funcs), m_handle(funcs.BadFileValue()) { }
		CloseFileScoper(FileOpFuncs &funcs, typename FileOpFuncs::FileHandleType fileHandle) : m_funcs(funcs), m_handle(fileHandle) { }
		~CloseFileScoper() { Close(); }
		inline bool                                 HasHandle() const                           { return !!m_funcs.IsGoodFileValue(m_handle); }
		inline void                                 Forget()                                    { m_handle = m_funcs.BadFileValue(); }
		inline bool                                 Close()                                     { if (!HasHandle()) return true; BOOL bRes = m_funcs.CloseFile(m_handle); Forget(); return !!bRes; }
		inline typename FileOpFuncs::FileHandleType Get() const                                 { assert(HasHandle()); return m_handle; }
		inline typename FileOpFuncs::FileHandleType GetAndForget()                              { assert(HasHandle()); typename FileOpFuncs::FileHandleType h = m_handle; Forget(); return m_handle; }
		inline void                                 Set(typename FileOpFuncs::FileHandleType h) { if (m_funcs.IsDifferentFileValue(h,m_handle)) { Close(); m_handle = h; } }

	};

	template < class FileOpFuncs >
	class DeleteFileScoper
	{
	private:
		FileOpFuncs &m_funcs;
		std::wstring m_strPath;
	private:
		DeleteFileScoper(const DeleteFileScoper &rhs); // disallow
		DeleteFileScoper &operator=(const DeleteFileScoper &rhs); // disallow
	public:
		DeleteFileScoper(FileOpFuncs &funcs) : m_funcs(funcs) { }
		DeleteFileScoper(FileOpFuncs &funcs, const wchar_t *szPath) : m_funcs(funcs), m_strPath(szPath) { }
		~DeleteFileScoper() { Delete(); }
		void Set(const wchar_t *szPath) { m_strPath = (szPath ? szPath : L""); }
		void Forget() { m_strPath.clear(); }
		bool HasFilePath() const { return !m_strPath.empty(); }
		void Delete() { if (HasFilePath()) { m_funcs.DeleteFile(m_strPath.c_str()); Forget(); } }
	};

	template < class FileOpFuncs >
	class RemoveDirectoryScoper
	{
	private:
		FileOpFuncs &m_funcs;
		std::wstring m_strPath;
	private:
		RemoveDirectoryScoper(const RemoveDirectoryScoper &rhs); // disallow
		RemoveDirectoryScoper &operator=(const RemoveDirectoryScoper &rhs); // disallow
	public:
		RemoveDirectoryScoper(FileOpFuncs &funcs) : m_funcs(funcs) { }
		RemoveDirectoryScoper(FileOpFuncs &funcs, const wchar_t *szPath) : m_funcs(funcs), m_strPath(szPath) { }
		~RemoveDirectoryScoper() { Remove(); }
		void Set(const wchar_t *szPath) { m_strPath = (szPath ? szPath : L""); }
		void Forget() { m_strPath.clear(); }
		bool HasDirPath() const { return !m_strPath.empty(); }
		void Remove() { if (HasDirPath()) { m_funcs.RemoveDirectory(m_strPath.c_str()); Forget(); } }
	};

	class ResumeThreadScoper
	{
	private:
		HANDLE m_hThread;
	private:
		ResumeThreadScoper(const ResumeThreadScoper &rhs); // disallow
		ResumeThreadScoper &operator=(const ResumeThreadScoper &rhs); // disallow
	public:
		ResumeThreadScoper(HANDLE hThread) : m_hThread(hThread) { }
		/* not virtual */ ~ResumeThreadScoper() { Resume(); }
		inline void Resume() { if (m_hThread != NULL) { ::ResumeThread(m_hThread); } m_hThread = NULL; }
	};

	class DestroyWindowScoper
	{
	private:
		HWND m_hWnd;
	private:
		DestroyWindowScoper(const DestroyWindowScoper &rhs) = delete;
		DestroyWindowScoper &operator=(const DestroyWindowScoper &rhs) = delete;
	public:
		DestroyWindowScoper() : m_hWnd(NULL) { }
		DestroyWindowScoper(HWND hWnd) : m_hWnd(hWnd) { }
		/* not virtual */ ~DestroyWindowScoper() { DestroyWindow(); }
		inline void DestroyWindow() { if (m_hWnd != NULL) { ::DestroyWindow(m_hWnd); } m_hWnd = NULL; }
		inline HWND Wnd() { return m_hWnd; }
		inline HWND operator=(HWND h) { if (m_hWnd != h) { DestroyWindow(); m_hWnd = h; } return h; }
	};

	class KillTimerScoper
	{
	private:
		HWND m_hWnd;
		UINT_PTR m_idTimer;
	private:
		KillTimerScoper(const KillTimerScoper &rhs); // disallow
		KillTimerScoper&operator=(const KillTimerScoper &rhs); // disallow
	public:
		KillTimerScoper(HWND hWnd, UINT_PTR idTimer) : m_hWnd(hWnd), m_idTimer(idTimer) { }
		KillTimerScoper(UINT_PTR idTimer) : m_hWnd(nullptr), m_idTimer(idTimer) { }
		/* not virtual */ ~KillTimerScoper() { KillTimer(); }
		inline void KillTimer() { if (m_idTimer != 0) { ::KillTimer(m_hWnd, m_idTimer); } m_hWnd = NULL; m_idTimer = 0; }
		inline operator UINT_PTR() const { return m_idTimer; }
		inline UINT_PTR Id() { return m_idTimer; }
	};

	class WindowDCScoper
	{
	private:
		HWND m_hWnd;
		HDC m_hDC;
	private:
		WindowDCScoper(const WindowDCScoper &rhs); // disallow
		WindowDCScoper &operator=(const WindowDCScoper &rhs); // disallow
	public:
		WindowDCScoper(HWND hWnd) : m_hWnd(hWnd), m_hDC(NULL) { /*if (m_hWnd != NULL)*/ { m_hDC = ::GetDC(m_hWnd); } }
		/* not virtual */ ~WindowDCScoper() { ReleaseDC(); }
		inline void ReleaseDC() { if (m_hDC != NULL) { ::ReleaseDC(m_hWnd, m_hDC); } m_hDC = NULL; m_hWnd = NULL; }
		inline HWND Wnd() { return m_hWnd; }
		inline HDC DC() { return m_hDC; }
	};

	class CompatibleDCScoper
	{
	private:
		HDC m_hDC;
	private:
		CompatibleDCScoper(const CompatibleDCScoper &rhs); // disallow
		CompatibleDCScoper &operator=(const CompatibleDCScoper &rhs); // disallow
	public:
		CompatibleDCScoper(HDC hDC) : m_hDC(NULL) { m_hDC = ::CreateCompatibleDC(hDC); }
		/* not virtual */ ~CompatibleDCScoper() { DeleteDC(); }
		inline void DeleteDC() { if (m_hDC != NULL) { ::DeleteDC(m_hDC); } m_hDC = NULL; }
		inline HDC DC() { return m_hDC; }
	};

	class SelectObjectScoper
	{
	private:
		HDC     m_hDC;
		HGDIOBJ m_hOldObj;
	private:
		SelectObjectScoper(const SelectObjectScoper &rhs) = delete;
		SelectObjectScoper &operator=(const SelectObjectScoper &rhs) = delete;
	public:
		SelectObjectScoper(HDC hDC, HGDIOBJ hNewObj) : m_hDC(hDC), m_hOldObj(::SelectObject(hDC, hNewObj)) { }
		/* not virtual */ ~SelectObjectScoper() { RestoreObject(); }
		inline void RestoreObject() { if (m_hOldObj != NULL) { ::SelectObject(m_hDC, m_hOldObj); m_hOldObj = NULL; } }
		inline HDC DC() { return m_hDC; }
	};

	class HBitmapScoper
	{
	private:
		HBITMAP m_hbm;
	private:
		HBitmapScoper(const HBitmapScoper &rhs) = delete;
		HBitmapScoper &operator=(const HBitmapScoper &rhs) = delete;
	public:
		HBitmapScoper(HBITMAP hbm) : m_hbm(hbm) { }
		/* not virtual */ ~HBitmapScoper() { FreeBitmap(); }
		inline void FreeBitmap() { if (m_hbm != NULL) { ::DeleteObject(m_hbm); m_hbm = NULL; } }
		inline HBITMAP HBM() { return m_hbm; }
		inline void Forget() { m_hbm = NULL; }
		inline HBITMAP GetAndForget() { HBITMAP h = m_hbm; m_hbm = NULL; return h; }
	};

	// ReleaseScoper may be used on things which are not COM objects.
	// The only assumption it makes about T is that it has AddRef() and Release() methods.
	// (The QueryInterfaceFrom and CreateInstance methods require more, obviously, but only if actually used.)
	template< typename T > class ReleaseScoper
	{
	private:
		T *m_p;
	public:
		ReleaseScoper()                         : m_p(NULL)                 { }
		ReleaseScoper(bool bAddRef, T *p)       : m_p(p)                    { if (bAddRef && m_p != NULL) { m_p->AddRef(); } }
		ReleaseScoper(const ReleaseScoper &rhs) : m_p(rhs.m_p)              { if (           m_p != NULL) { m_p->AddRef(); } }
		ReleaseScoper(ReleaseScoper &&rhs)      : m_p( std::move(rhs.m_p) ) { rhs.m_p = NULL; }
		ReleaseScoper &operator=(const ReleaseScoper &rhs)                  { SetWithAddRef(rhs.m_p); return *this; }
		ReleaseScoper &operator=(ReleaseScoper &&rhs)                       { if (this != &rhs) { SetWithoutAddRef(rhs.m_p); rhs.m_p = NULL; } return *this; }
		/* not virtual */ ~ReleaseScoper()                                  { ReleaseChild(); }
		inline bool IsEmpty() const                                         { return (m_p == NULL); }
		inline void Forget()                                                { m_p = NULL; }
		inline void ReleaseChild()                                          { if (m_p != NULL) { m_p->Release(); Forget(); } }
		inline const T * Get() const                                        { return m_p; }
		inline T * GetNonConstExplicit() const                              { return m_p; } // If the scoper is const then it usually makes sense for the object within it to be as well, but not always. In the exceptional cases, callers have to be explicit.
		inline T * Get()                                                    { return m_p; }
		inline T * GetAndForget()                                           { T *p = m_p; Forget(); return p; }
		inline T * * AssignableRelease()                                    { ReleaseChild(); return &m_p; }
		inline void SetWithoutAddRef(T *p)                                  { ReleaseChild(); m_p = p; } // If m_p==p already then this effectively releases one reference on p. That's correct; someone else is giving us their reference & we can release our old one.
		inline void SetWithAddRef(T *p)                                     { if (p != NULL) { p->AddRef(); } SetWithoutAddRef(p); } // Correctly handles case where p is already what we store.
		inline REFIID UUID() const                                          { return __uuidof(T); }

		inline bool QueryInterfaceFrom(IUnknown *pUnk, const IID &riid) // Any existing child will be released even if this fails.
		{
			// Correctly handles case where pUnk is m_p (i.e. we must not release it until after the QI call).

			T *p = nullptr;

			if (pUnk == NULL || pUnk->QueryInterface(riid, reinterpret_cast< void ** >(&p)) != S_OK || p == NULL)
			{
				ReleaseChild();
				return false;
			}

			SetWithoutAddRef(p);
			return true;
		}

		inline bool QueryInterfaceFrom(IUnknown *pUnk) { return QueryInterfaceFrom(pUnk, __uuidof(T)); } // Any existing child will be released even if this fails.

		inline bool CreateInstance(const CLSID &clsid, const IID &riid) // Any existing child will be released (before the attempt) even if this fails.
		{
			ReleaseChild();

			T *p = nullptr;
			
			if (::CoCreateInstance(clsid, NULL, CLSCTX_INPROC_SERVER|CLSCTX_LOCAL_SERVER, riid, reinterpret_cast< void ** >(&p)) != S_OK || p == NULL)
			{
				return false;
			}

			SetWithoutAddRef(p);
			return true;
		}

		inline bool CreateInstance(const CLSID &clsid) { return CreateInstance(clsid, __uuidof(T)); } // Any existing child will be released (before the attempt) even if this fails.
	};

	template <typename T> class RefCounter
	{
	private:
		T *           m_p;
		volatile LONG m_lRefCount;
	private:
		RefCounter(const RefCounter &rhs); // disallow
		RefCounter &operator=(const RefCounter &rhs); // disallow
		~RefCounter() { delete m_p; m_p = 0; } // non-virtual
	public:
		RefCounter(T *p) : m_p(p), m_lRefCount(1) { }
		      T *Get()       { return m_p; }
		const T *Get() const { return m_p; }
		LONG AddRef()  { return ::InterlockedIncrement(&m_lRefCount); };
		LONG Release() { LONG lRefCount = ::InterlockedDecrement(&m_lRefCount); if (lRefCount == 0) { delete this; } return lRefCount; };
	};

	// Never pass DeleteScoper objects across process/DLL boundaries.
	// (It's all inline so the new/delete calls may hit different heaps.)
	// Do not use DeleteScoper with arrays that need delete[]; use ArrayScoper for those.
	template <typename T> class DeleteScoper
	{
	protected:
		T *m_p;
	private:
		DeleteScoper(const DeleteScoper &rhs); // disallow
		DeleteScoper &operator=(const DeleteScoper &rhs); // disallow
	public:
		DeleteScoper() : m_p(NULL)        { }
		DeleteScoper(T *p) : m_p(p)       { }
		/* not virtual */ ~DeleteScoper() { Free(); }
		inline bool     IsEmpty() const   { return (m_p == NULL); }
		inline void     Free()            { delete m_p; m_p = NULL; }
		inline void     Forget()          { m_p = NULL; }
		inline void     Set(T *p)         { if (p != m_p) { Free(); m_p = p; } }
		inline       T *Get()             { return m_p; }
		inline const T *Get() const       { return m_p; }
		inline       T *GetAndForget()    { T *p = m_p; Forget(); return p; }
	//	operator T*() { return m_p; }
	};

	class BStrFreer
	{
	protected:
		BSTR m_bstr;
	private:
		BStrFreer(const BStrFreer &rhs); // disallow
		BStrFreer &operator=(const BStrFreer &rhs); // disallow
	public:
		BStrFreer()          : m_bstr(NULL)           { }
		BStrFreer(BSTR bstr) : m_bstr(bstr)           { }
		/* not virtual */ ~BStrFreer()                { Free(); }
		      bool IsEmpty() const                    { return (m_bstr != NULL); }
		      void Free()                             { ::SysFreeString(m_bstr); m_bstr = NULL; }
		      void Forget()                           { m_bstr = NULL; }
			  void TakeBstr(BSTR bstr)                { if (bstr != m_bstr) { Free(); m_bstr = bstr; } }
		      void AllocFromString(const wchar_t *sz) { Free(); m_bstr = ::SysAllocString(sz); }
		      BSTR Get()                              { return m_bstr; }
		const BSTR Get() const                        { return m_bstr; }
		      BSTR GetAndForget()                     { BSTR bstr = m_bstr; m_bstr = NULL; return bstr; }
	};

	class BStrMaker
	{
	protected:
		BSTR m_bstr;
	private:
		BStrMaker(const BStrMaker &rhs); // disallow
		BStrMaker &operator=(const BStrMaker &rhs); // disallow
	public:
		BStrMaker(const wchar_t *sz) : m_bstr(::SysAllocString(sz)) { }
		~BStrMaker() { ::SysFreeString(m_bstr); m_bstr = NULL; }
		BSTR Get() const { return m_bstr; }
	};

	class BStrMakerPreserveErrors // Special version that preserves last-error during construction/destruction.
	{
	protected:
		BSTR m_bstr;
	private:
		BStrMakerPreserveErrors(const BStrMakerPreserveErrors &rhs); // disallow
		BStrMakerPreserveErrors &operator=(const BStrMakerPreserveErrors &rhs); // disallow
	public:
		BStrMakerPreserveErrors(const wchar_t *sz) : m_bstr(NULL)
		{
			DWORD dwErr = ::GetLastError();
			m_bstr = ::SysAllocString(sz);
			::SetLastError(dwErr);
		}
		~BStrMakerPreserveErrors() 
		{
			DWORD dwErr = ::GetLastError();
			::SysFreeString(m_bstr);
			m_bstr = NULL;
			::SetLastError(dwErr);
		}
		BSTR Get() const { return m_bstr; }
	};

	// Never pass ArrayScoper objects across process/DLL boundaries.
	// (It's all inline so the new/delete calls may hit different heaps.)
	template <typename T> class ArrayScoper
	{
	protected:
		T *m_p;
	private:
		ArrayScoper(const ArrayScoper &rhs); // disallow
		ArrayScoper &operator=(const ArrayScoper &rhs); // disallow
	public:
		ArrayScoper() : m_p(NULL)        { }
		ArrayScoper(T *p) : m_p(p)       { }
		/* not virtual */ ~ArrayScoper() { Free(); }
		inline bool     IsEmpty() const  { return (m_p == NULL); }
		inline void     Free()           { delete[] m_p; m_p = NULL; }
		inline void     Forget()         { m_p = NULL; }
		inline void     Set(T *p)        { if (p != m_p) { Free(); m_p = p; } }
		inline       T *Get()            { return m_p; }
		inline const T *Get() const      { return m_p; }
		inline       T *GetAndForget()   { T *p = m_p; Forget(); return p; }
		inline      T **GetAssignment()  { Free(); return &m_p; }
	//	operator T*() { return m_p; }
	};

	typedef ArrayScoper< wchar_t > WCharArrayScoper;

	/*
	// ArrayScoperDelOpt is like ArrayScoper but it can also be given static arrays that aren't for deletion.
	// Never pass ArrayScoperDelOpt objects across process/DLL boundaries.
	// (It's all inline so the new/delete calls may hit different heaps.)
	template <typename T> class ArrayScoperDelOpt
	{
	protected:
		T *m_p;
		bool m_bDelete;
	private:
		ArrayScoperDelOpt(const ArrayScoperDelOpt &rhs); // disallow
		ArrayScoperDelOpt &operator=(const ArrayScoperDelOpt &rhs); // disallow
	public:
		ArrayScoperDelOpt()                   : m_p(NULL), m_bDelete(false)   { }
		ArrayScoperDelOpt(bool bDelete, T *p) : m_p(p),    m_bDelete(bDelete) { }
		~ArrayScoperDelOpt()                           { Free(); } // non-virtual
		inline bool     IsEmpty() const                { return (m_p == NULL); }
		inline bool     IsGoignToDelete() const        { return m_bDelete; }
		inline void     Free()                         { if (m_bDelete) { delete[] m_p; } m_p = NULL; m_bDelete = false; }
		inline void     Forget()                       { m_p = NULL; m_bDelete = false; }
		inline void     Set(bool bDelete, T *p)        { if (p != m_p) { Free(); m_p = p; m_bDelete = bDelete; } else if (bDelete) { m_bDelete = true; } }
		inline       T *Get()            { return m_p; }
		inline const T *Get() const      { return m_p; }
		inline       T *GetAndForget()   { T *p = m_p; Forget(); return p; }
	//	operator T*() { return m_p; }
	};

	typedef ArrayScoperDelOpt< wchar_t > WCharArrayScoperDelOpt;
	*/

	// Never pass MultiArrayScoper objects across process/DLL boundaries.
	// (It's all inline so the new/delete calls may hit different heaps.)
	class MultiArrayScoper
	{
	private:
		std::vector< void * > m_voidPointers;
	private:
		MultiArrayScoper(const MultiArrayScoper &rhs); // disallow
		MultiArrayScoper &operator=(const MultiArrayScoper &rhs); // disallow
	public:
		MultiArrayScoper() { }
		/* not virtual */ ~MultiArrayScoper() { while(!m_voidPointers.empty()) { delete [] m_voidPointers.back(); m_voidPointers.pop_back(); } }
		void Add(void *pVoid) { m_voidPointers.push_back(pVoid); }
	};

	// Never pass AutoBuffer objects across process/DLL boundaries.
	// (It's all inline so the new/delete calls may hit different heaps.)
	template <typename T> class AutoBuffer
	{
	private:
		BYTE *m_pBuffer;
		size_t m_size;
		size_t m_sizeInUse; // May be less than m_size if AllocateMinimum* or TruncateSizeInUse were used.
	private:
		AutoBuffer(const AutoBuffer &rhs) = delete;
		AutoBuffer &operator=(const AutoBuffer &rhs) = delete;
	public:
		AutoBuffer() : m_pBuffer(0) , m_size(0), m_sizeInUse(0) { }
		/* not virtual */ ~AutoBuffer() { Free(); } // Must not change Win32 LastError
		void Free() { delete[] m_pBuffer; m_pBuffer = 0; m_size = 0; m_sizeInUse = 0; } // Must not change Win32 LastError

		void Swap(AutoBuffer &rhs)
		{
			BYTE *pBuffer    = rhs.m_pBuffer;
			size_t size      = rhs.m_size;
			size_t sizeInUse = rhs.m_sizeInUse;

			rhs.m_pBuffer   = m_pBuffer;
			rhs.m_size      = m_size;
			rhs.m_sizeInUse = m_sizeInUse;

			m_pBuffer   = pBuffer;
			m_size      = size;
			m_sizeInUse = sizeInUse;
		}

		// Allocating a zero-length buffer is legal and results in a non-null buffer.
		// Make no assumptions about the content of the buffer after allocation.
		bool AllocateBytes(size_t s)
		{
			if (s == m_size && m_pBuffer != 0) { return true; }
			Free();
			m_pBuffer = new(std::nothrow) BYTE[s];
			if (m_pBuffer == 0) { m_size = 0; m_sizeInUse = 0; ::SetLastError(ERROR_NOT_ENOUGH_MEMORY); return false; }
			m_size = s;
			m_sizeInUse = s;
			return true;
		}

		// If there's already a buffer big enough (or bigger) then do nothing, else (re-)allocate.
		bool AllocateMinimumBytes(size_t s)
		{
			if (s <= m_size && m_pBuffer != 0) { m_sizeInUse = s; return true; }
			return AllocateBytes(s);
		}

		bool        AllocateElements(size_t c) {        return AllocateBytes(c * sizeof(T)); }
		bool AllocateMinimumElements(size_t c) { return AllocateMinimumBytes(c * sizeof(T)); }

		T *       GetBuffer()                  { return reinterpret_cast<       T * >(m_pBuffer); }
		const T * GetBuffer()            const { return reinterpret_cast< const T * >(m_pBuffer); }
		size_t    GetSizeBytes()         const { return m_size;                    }
		size_t    GetSizeBytesInUse()    const { return m_sizeInUse;               }
		size_t    GetSizeElements()      const { return (m_size / sizeof(T));      }
		size_t    GetSizeElementsInUse() const { return (m_sizeInUse / sizeof(T)); }
		bool      IsAllocated()          const { return (m_pBuffer != 0);          }
		bool      TruncateSizeInUse(size_t s)  { if (m_sizeInUse >= s) { m_sizeInUse = s; return true; } return false; }
	};

	template <typename T> class FreeAutoBufferOnDestruction
	{
	private:
		AutoBuffer<T> *m_pBuf{nullptr};
	private:
		FreeAutoBufferOnDestruction(const FreeAutoBufferOnDestruction &rhs) = delete;
		FreeAutoBufferOnDestruction &operator=(const FreeAutoBufferOnDestruction &rhs) = delete;
	public:
		FreeAutoBufferOnDestruction(AutoBuffer<T> &buf) : m_pBuf(&buf) { }
		~FreeAutoBufferOnDestruction()                                 { FreeNowAndForget(); }
		void FreeNowAndForget()                                        { if (m_pBuf) { m_pBuf->Free(); Forget(); } }
		void Forget()                                                  { m_pBuf = nullptr; }
	};

	template < typename T > class MallocFreeScoper
	{
	private:
		void *m_pBuffer;
		size_t m_size;
		size_t m_sizeInUse;
	private:
		MallocFreeScoper(const MallocFreeScoper &rhs); // disallow
		MallocFreeScoper &operator=(const MallocFreeScoper &rhs); // disallow
	public:
		MallocFreeScoper() : m_pBuffer(0), m_size(0), m_sizeInUse(0) { }
		/* not virtual */ ~MallocFreeScoper()   { Free(); } // Must not change Win32 LastError
		void Forget()                    { m_pBuffer = NULL; m_size = 0; m_sizeInUse = 0; }
		void Free()                      { if (m_pBuffer != 0) { ::free(m_pBuffer); Forget(); } }

		void Set(void* pBuffer, size_t size)
		{
			Free();
			m_pBuffer = pBuffer;
			m_size = size;
			m_sizeInUse = size;
		}
		
		// Allocating a zero-length buffer is legal and results in a non-null buffer.
		// Make no assumptions about the content of the buffer after allocation.
		bool AllocateBytes(size_t s)
		{
			if (s == m_size && m_pBuffer != 0) { return true; }
			Free();
			m_pBuffer = ::malloc(s);
			if (m_pBuffer == 0) { m_size = 0; m_sizeInUse = 0; ::SetLastError(ERROR_NOT_ENOUGH_MEMORY); return false; }
			m_size = s;
			m_sizeInUse = s;
			return true;
		}

		// If there's already a buffer big enough (or bigger) then do nothing, else (re-)allocate.
		bool AllocateMinimumBytes(size_t s)
		{
			if (s <= m_size && m_pBuffer != 0) { m_sizeInUse = s; return true; }
			return AllocateBytes(s);
		}

		bool        AllocateElements(size_t c) {        return AllocateBytes(c * sizeof(T)); }
		bool AllocateMinimumElements(size_t c) { return AllocateMinimumBytes(c * sizeof(T)); }

		T *       GetBuffer()                  { return reinterpret_cast<       T * >(m_pBuffer); }
		const T * GetBuffer()            const { return reinterpret_cast< const T * >(m_pBuffer); }
		T *       GetBufferAndForget()         { T *pBuffer = GetBuffer(); Forget(); return pBuffer; }
		size_t    GetSizeBytes()         const { return m_size;                    }
		size_t    GetSizeBytesInUse()    const { return m_sizeInUse;               }
		size_t    GetSizeElements()      const { return (m_size / sizeof(T));      }
		size_t    GetSizeElementsInUse() const { return (m_sizeInUse / sizeof(T)); }
		bool      IsAllocated()          const { return (m_pBuffer != 0);          }
	};

	class LocalFreeScoper
	{
	private:
		HLOCAL m_h;
		bool m_bHasHandle;
	private:
		LocalFreeScoper(const LocalFreeScoper &rhs); // disallow
		LocalFreeScoper &operator=(const LocalFreeScoper &rhs); // disallow
	public:
		LocalFreeScoper()         : m_h(NULL), m_bHasHandle(false) { }
		LocalFreeScoper(HLOCAL h) : m_h(h),    m_bHasHandle(true)  { }
		/* not virtual */ ~LocalFreeScoper() { Free(); }
		inline bool HasHandle() const         { return m_bHasHandle; }
		inline void Forget()                  { m_h = NULL; m_bHasHandle = false; }
		inline void Free()                    { if (m_bHasHandle) { ::LocalFree(m_h); Forget(); } }
		inline HLOCAL Get() const             { assert(m_bHasHandle); return m_h; }
		inline HLOCAL GetAndForget()          { HLOCAL h = m_h; Forget(); return h; }
		inline void Set(HLOCAL h)             { if (h != m_h) { Free(); m_h = h; m_bHasHandle = true; } }
		inline bool AllocateFixed(SIZE_T uBytes)
		{
			Free();
			m_h = ::LocalAlloc(LMEM_FIXED, uBytes);
			if (m_h) { m_bHasHandle = true; }
			return m_bHasHandle;
		}
	};

	class CryptProviderScoper
	{
	private:
		HCRYPTPROV m_h;
		bool m_bHasHandle;
	private:
		CryptProviderScoper(const CryptProviderScoper &rhs) = delete;
		CryptProviderScoper &operator=(const CryptProviderScoper &rhs) = delete;
	public:
		CryptProviderScoper()             : m_h(NULL), m_bHasHandle(false)   { }
		CryptProviderScoper(HCRYPTPROV h) : m_h(h),    m_bHasHandle(h!=NULL) { }
		/* not virtual */ ~CryptProviderScoper() { Free(); }
		inline bool Aquire(LPCWSTR szContainer, LPCWSTR szProvider, DWORD dwProvType, DWORD dwFlags)
		{
			Free(); 
			if (!::CryptAcquireContext(&m_h, szContainer, szProvider, dwProvType, dwFlags))
			{
				return false;
			}
			m_bHasHandle = true;
			return true;
		}
		inline bool HasHandle() const            { return m_bHasHandle; }
		inline void Forget()                     { m_h = NULL; m_bHasHandle = false; }
		inline void Free()                       { if (m_bHasHandle) { ::CryptReleaseContext(m_h, 0); Forget(); } }
		inline HCRYPTPROV Get() const            { assert(m_bHasHandle); return m_h; }
		inline HCRYPTPROV GetAndForget()         { HCRYPTPROV h = m_h; Forget(); return h; }
		inline void Set(HCRYPTPROV h)            { if (h != m_h) { Free(); m_h = h; m_bHasHandle = true; } }
	};

	class CryptHashScoper
	{
	private:
		HCRYPTHASH m_h;
		bool m_bHasHandle;
	private:
		CryptHashScoper(const CryptHashScoper &rhs) = delete;
		CryptHashScoper &operator=(const CryptHashScoper &rhs) = delete;
	public:
		CryptHashScoper()             : m_h(NULL), m_bHasHandle(false)   { }
		CryptHashScoper(HCRYPTHASH h) : m_h(h),    m_bHasHandle(h!=NULL) { }
		/* not virtual */ ~CryptHashScoper() { Free(); }
		inline bool Create(HCRYPTPROV hProv, ALG_ID Algid, HCRYPTKEY hKey, DWORD dwFlags)
		{
			Free(); 
			if (!::CryptCreateHash(hProv, Algid, hKey, dwFlags, &m_h))
			{
				return false;
			}
			m_bHasHandle = true;
			return true;
		}
		inline bool HasHandle() const            { return m_bHasHandle; }
		inline void Forget()                     { m_h = NULL; m_bHasHandle = false; }
		inline void Free()                       { if (m_bHasHandle) { ::CryptDestroyHash(m_h); Forget(); } }
		inline HCRYPTHASH Get() const            { assert(m_bHasHandle); return m_h; }
		inline HCRYPTHASH GetAndForget()         { HCRYPTHASH h = m_h; Forget(); return h; }
		inline void Set(HCRYPTHASH h)            { if (h != m_h) { Free(); m_h = h; m_bHasHandle = true; } }
	};

	class CryptKeyScoper
	{
	private:
		HCRYPTKEY m_h;
		bool m_bHasHandle;
	private:
		CryptKeyScoper(const CryptKeyScoper &rhs) = delete;
		CryptKeyScoper &operator=(const CryptKeyScoper &rhs) = delete;
	public:
		CryptKeyScoper()            : m_h(NULL), m_bHasHandle(false)   { }
		CryptKeyScoper(HCRYPTKEY h) : m_h(h),    m_bHasHandle(h!=NULL) { }
		/* not virtual */ ~CryptKeyScoper() { Free(); }
		inline bool Derive(HCRYPTPROV hProv, ALG_ID Algid, HCRYPTHASH hBaseData, DWORD dwFlags)
		{
			Free(); 
			if (!::CryptDeriveKey(hProv, Algid, hBaseData, dwFlags, &m_h))
			{
				return false;
			}
			m_bHasHandle = true;
			return true;
		}
		inline bool HasHandle() const            { return m_bHasHandle; }
		inline void Forget()                     { m_h = NULL; m_bHasHandle = false; }
		inline void Free()                       { if (m_bHasHandle) { ::CryptDestroyKey(m_h); Forget(); } }
		inline HCRYPTKEY Get() const             { assert(m_bHasHandle); return m_h; }
		inline HCRYPTKEY GetAndForget()          { HCRYPTKEY h = m_h; Forget(); return h; }
		inline void Set(HCRYPTKEY h)             { if (h != m_h) { Free(); m_h = h; m_bHasHandle = true; } }
	};

	class PasswordCrypter
	{
	private:
		LeoHelpers::CryptProviderScoper m_scCryptProvider;
		LeoHelpers::CryptHashScoper     m_scCryptHash;
		LeoHelpers::CryptKeyScoper      m_scCryptKey;
		bool m_fInitOk { false };

	public:
		PasswordCrypter(const wchar_t *szPassword)
		{
			if (!szPassword || !*szPassword)
			{
				::SetLastError(ERROR_INVALID_PARAMETER);
				return;
			}

			size_t lenPasswordBytes = wcslen(szPassword) * sizeof(szPassword[0]);

			if (lenPasswordBytes > 0xffffffffUL)
			{
				::SetLastError(ERROR_MESSAGE_EXCEEDS_MAX_SIZE);
				return;
			}

			if (!m_scCryptProvider.Aquire(NULL, NULL, PROV_RSA_FULL, CRYPT_VERIFYCONTEXT|CRYPT_SILENT))
			{
				return;
			}

			if (!m_scCryptHash.Create(m_scCryptProvider.Get(), CALG_SHA,  NULL, 0))
			{
				return;
			}

			if (!::CryptHashData(m_scCryptHash.Get(), reinterpret_cast< const BYTE * >( szPassword ), static_cast< DWORD >( lenPasswordBytes ), 0))
			{
				return;
			}


			if (!m_scCryptKey.Derive(m_scCryptProvider.Get(), CALG_RC4, m_scCryptHash.Get(), CRYPT_EXPORTABLE|CRYPT_NO_SALT))
			{
				return;
			}

			m_fInitOk = true;
		}

		bool InitOk() const { return m_fInitOk; }

		bool Encrypt(const void *pData, size_t dataSize, LeoHelpers::AutoBuffer< BYTE > &scResult)
		{
			if (!InitOk())
			{
				::SetLastError(ERROR_INVALID_HANDLE_STATE);
				return false;
			}

			if (!pData || dataSize == 0)
			{
				::SetLastError(ERROR_INVALID_PARAMETER);
				return false;
			}

			if (dataSize > 0xffffffffUL)
			{
				::SetLastError(ERROR_MESSAGE_EXCEEDS_MAX_SIZE);
				return false;
			}

			DWORD dwDataLenRequired = static_cast< DWORD >( dataSize );

			if (!::CryptEncrypt(m_scCryptKey.Get(), NULL, TRUE, 0, NULL, &dwDataLenRequired, 0) || dwDataLenRequired == 0)
			{
				return false;
			}

			if (!scResult.AllocateMinimumBytes(dwDataLenRequired))
			{
				return false;
			}

			memcpy_s(scResult.GetBuffer(), scResult.GetSizeBytes(), pData, dataSize);

			dwDataLenRequired = static_cast< DWORD >( dataSize );

			if (!::CryptEncrypt(m_scCryptKey.Get(), NULL, TRUE, 0, scResult.GetBuffer(), &dwDataLenRequired, static_cast< DWORD >( scResult.GetSizeBytesInUse() )))
			{
				return false;
			}

			return true;
		}

		bool Decrypt(const void *pData, size_t dataSize, LeoHelpers::AutoBuffer< BYTE > &scResult)
		{
			if (!InitOk())
			{
				::SetLastError(ERROR_INVALID_HANDLE_STATE);
				return false;
			}

			if (!pData || dataSize == 0)
			{
				::SetLastError(ERROR_INVALID_PARAMETER);
				return false;
			}

			if (dataSize > 0xffffffffUL)
			{
				::SetLastError(ERROR_MESSAGE_EXCEEDS_MAX_SIZE);
				return false;
			}

			DWORD dwDataLenRequired = static_cast< DWORD >( dataSize );

			if (!scResult.AllocateMinimumBytes(dwDataLenRequired))
			{
				return false;
			}

			memcpy_s(scResult.GetBuffer(), scResult.GetSizeBytes(), pData, dataSize);

			dwDataLenRequired = static_cast< DWORD >( dataSize );

			if (!::CryptDecrypt(m_scCryptKey.Get(), NULL, TRUE, 0, scResult.GetBuffer(), &dwDataLenRequired) || dwDataLenRequired == 0)
			{
				return false;
			}

			if (!scResult.TruncateSizeInUse(dwDataLenRequired))
			{
				return false;
			}

			return true;
		}
	};

	class BinaryData
	{
	protected:
		BYTE *m_pData;
		DWORD m_dwSize;
		std::shared_ptr<std::vector<BYTE>> m_pVec;
		bool  m_bUseLocalAlloc;
		bool  m_bUseSharedVector;

	public:
		BinaryData(bool bUseLocalAlloc, bool bUseSharedVector)
		: m_pData(NULL)
		, m_dwSize(0)
		, m_bUseLocalAlloc(bUseLocalAlloc)
		, m_bUseSharedVector(bUseSharedVector)
		{
			assert(!(bUseLocalAlloc && bUseSharedVector));
		}

		~BinaryData()
		{
			Clear();
		}

	private:
		BinaryData(const BinaryData &rhs); // disallow
		BinaryData &operator=(const BinaryData &rhs); // disallow

	public:

		void Clear()
		{
			if (m_pData != NULL)
			{
				if (m_bUseSharedVector)
				{
					m_pVec.reset();
				}
				else
				if (m_bUseLocalAlloc)
				{
					::LocalFree(m_pData);
				}
				else
				{
					delete[] m_pData;
				}
			}
			m_pData = NULL;
			m_dwSize = 0;
		}

		bool AllocateBytes(DWORD dwByteSize)
		{
			Clear();

			if (m_bUseSharedVector)
			{
				m_pVec.reset(new std::vector<BYTE>(dwByteSize));
				m_pData = m_pVec->data();
			}
			else
			if (m_bUseLocalAlloc)
			{
				m_pData = reinterpret_cast< BYTE *>( ::LocalAlloc(LMEM_FIXED, dwByteSize) );
			}
			else
			{
				m_pData = new(std::nothrow) BYTE[dwByteSize];
			}

			if (m_pData == NULL)
			{
				::SetLastError(ERROR_NOT_ENOUGH_MEMORY);
				return false;
			}

			m_dwSize = dwByteSize;
			return true;
		}

		void Forget()
		{
			m_pData = NULL;
			m_dwSize = 0;
			m_pVec.reset();
		}

		void SetAndOwn(BYTE *pData, DWORD dwSize, bool bUseLocalAlloc)
		{
			Clear();

			m_pData = pData;
			m_dwSize = dwSize;
			m_bUseLocalAlloc = bUseLocalAlloc;
			m_bUseSharedVector = false;
		}

		// Use ::GetLastError on failure.
		// Give dwMaxRead=INVALID_FILE_SIZE to load the whole file.
		// If the file is larger than requested, only the requested amount is loaded.
		// If the file is smaller than requested, the full file is loaded and success is returned.
		// If the file is empty, returns success with an empty (zero-length) buffer.
		// Will not load files larger than 4GB (and this would be a stupid way to load such things).
		bool Load(const wchar_t *filename, DWORD dwMaxRead=INVALID_FILE_SIZE, DWORD dwOnlyReadIfExactSize=INVALID_FILE_SIZE)
		{
			bool bSuccess = false;
			DWORD dwLastError = ERROR_INTERNAL_ERROR;

			Clear();

			HANDLE hFile = ::CreateFileW(filename, GENERIC_READ, FILE_SHARE_READ, NULL, OPEN_EXISTING, FILE_FLAG_SEQUENTIAL_SCAN, NULL);

			if (INVALID_HANDLE_VALUE == hFile)
			{
				dwLastError = ::GetLastError();
			}
			else
			{
				DWORD dwFileSize = ::GetFileSize(hFile, NULL);

				if (INVALID_FILE_SIZE == dwFileSize && NO_ERROR != ::GetLastError())
				{
					dwLastError = ::GetLastError();
				}
				else if (dwOnlyReadIfExactSize != INVALID_FILE_SIZE && dwOnlyReadIfExactSize != dwFileSize)
				{
					dwLastError = ERROR_INCORRECT_SIZE;
				}
				else
				{
					if (dwMaxRead != INVALID_FILE_SIZE && dwFileSize > dwMaxRead)
					{
						dwFileSize = dwMaxRead; // File is bigger than the requested max, so only read the requested max.
					}

					if (!AllocateBytes(dwFileSize))
					{
						dwLastError = ::GetLastError();
					}
					else
					{
						if (0 == ::ReadFile(hFile, m_pData, dwFileSize, &m_dwSize, NULL))
						{
							dwLastError = ::GetLastError();
						}
						else if (dwOnlyReadIfExactSize != INVALID_FILE_SIZE && dwOnlyReadIfExactSize != m_dwSize)
						{
							dwLastError = ERROR_INCORRECT_SIZE;
						}
						else
						{
							dwLastError = ERROR_SUCCESS;
							bSuccess = true;
						}
					}
				}

				::CloseHandle(hFile);
			}

			if (bSuccess)
			{
				return true;
			}

			Clear();
			::SetLastError(dwLastError);
			return false;
		}

		bool Save(const wchar_t *filename)
		{
			return Save(filename, m_pData, m_dwSize);
		}

		static bool Save(const wchar_t *filename, const BYTE *pData, DWORD dwSize)
		{
			bool bSuccess = false;
			DWORD dwLastError = ERROR_INTERNAL_ERROR;

			HANDLE hFile = ::CreateFile(filename, FILE_WRITE_DATA, FILE_SHARE_READ, NULL, CREATE_NEW, FILE_FLAG_SEQUENTIAL_SCAN, NULL);

			if (hFile == INVALID_HANDLE_VALUE)
			{
				dwLastError = ::GetLastError();
			}
			else
			{
				DWORD dwWritten = 0;

				if (!::WriteFile(hFile, pData, dwSize, &dwWritten, NULL))
				{
					dwLastError = ::GetLastError();
				}
				else
				{
					if (dwWritten != dwSize)
					{
						dwLastError = ERROR_WRITE_FAULT;
					}
					else
					{
						bSuccess = true;
					}
				}

				::CloseHandle(hFile);

				if (!bSuccess)
				{
					::DeleteFile(filename);
				}
			}

			::SetLastError(dwLastError);
			return bSuccess;
		}

		BYTE *GetBuffer()
		{
			return m_pData;
		}

		const BYTE *GetBuffer() const
		{
			return m_pData;
		}

		DWORD GetSize() const
		{
			return m_dwSize;
		}

		std::shared_ptr<std::vector<BYTE>> GetSharedVector()
		{
			if (m_pVec.get())
				return m_pVec;

			return std::make_shared<std::vector<BYTE>>(m_pData, m_pData + m_dwSize);
		}

		const BYTE * Contains(const BYTE *pSearch, DWORD dwSearchLen) const
		{
			if (!m_pData || m_dwSize < dwSearchLen || !pSearch)
			{
				return nullptr;
			}

			if (dwSearchLen == 0)
			{
				return m_pData;
			}

			const BYTE *pDataEnd = m_pData + (m_dwSize - dwSearchLen);
			const BYTE *pSearchEnd = pSearch + dwSearchLen;

			for (const BYTE *pData = m_pData; pData <= pDataEnd; ++pData)
			{
				const BYTE *pS = pSearch;
				const BYTE *pD = pData;
				while (true)
				{
					if (pS == pSearchEnd)
					{
						return pData;
					}

					if (*pD++ != *pS++)
					{
						break;
					}
				}
			}

			return nullptr;
		}
	};

	// Does not copy the resource; the module it comes from must remain loaded while the data is used.
	class BinaryResource
	{
		const void * m_pData{};
		DWORD m_dwSize{};
	public:
		BinaryResource(const BinaryResource &rhs) = default;
		BinaryResource &operator=(const BinaryResource &rhs) = default;

		BinaryResource(HINSTANCE hInstance, LPCWSTR lpszName, LPCWSTR lpszType)
		{
			HRSRC hResource = FindResource(hInstance, lpszName, lpszType);
			if (hResource)
			{
				HGLOBAL hGlobal = LoadResource(hInstance, hResource);

				if (hGlobal)
				{
					m_pData = LockResource(hGlobal);

					if (m_pData)
					{
						m_dwSize = SizeofResource(hInstance, hResource);
					}
				}
			}
		}

		~BinaryResource()
		{
			m_pData = nullptr;
			m_dwSize = 0;
		}

		const void * GetBuffer() const
		{
			return m_pData;
		}

		DWORD GetSize() const
		{
			return m_dwSize;
		}
	};

	// Ref-counted ID object which can outlive the parent object it represents. Similar to comparing the address
	// of a the parent object instance but without the problem that if that parent object is destroyed and another
	// created, the new one may have the same address as the old and look like the same object to anything that
	// compared a stored (now dangling/recycled) pointer. Since the ID object is ref-counted, its address won't be
	// recycled until all interested parties have released it. This also removes problems with circular references
	// since the InstanceIdent itself never references any other object.
	class InstanceIdent
	{
	private:
		volatile LONG m_lRefCount;

		InstanceIdent(const InstanceIdent &rhs); // disallow
		InstanceIdent &operator=(const InstanceIdent &rhs); // disallow

		// Destructor private as destruction must be through Release.
		~InstanceIdent() { } // Non-virtual.
	public:
		InstanceIdent() : m_lRefCount(1) { }
		ULONG AddRef()  { return InterlockedIncrement(&m_lRefCount); }
		ULONG Release() { LONG lRefCount = InterlockedDecrement(&m_lRefCount); if (lRefCount==0) { delete this; } return lRefCount; }

	//	bool operator==(const InstanceIdent &rhs)     const { return (this==&rhs); }
	//	bool operator!=(const InstanceIdent &rhs)     const { return (this!=&rhs); }
	//	bool IsEqualTo(const InstanceIdent &rhs)      const { return (this==&rhs); }
		bool IsEqualTo(const InstanceIdent *pRhs)     const { return (this==pRhs); }
	//	bool IsDifferentTo(const InstanceIdent &rhs)  const { return (this!=&rhs); }
		bool IsDifferentTo(const InstanceIdent *pRhs) const { return (this!=pRhs); }
	};

	/*
	class EnvBlockWideScoper
	{
	private:
		LPWCH m_envStringsW;
	private:
		EnvBlockWideScoper(const EnvBlockWideScoper &rhs); // disallow
		EnvBlockWideScoper &operator=(const EnvBlockWideScoper &rhs); // disallow
	public:
		EnvBlockWideScoper() : m_envStringsW(::GetEnvironmentStringsW()) { }
		~EnvBlockWideScoper()  { ::FreeEnvironmentStringsW(m_envStringsW); m_envStringsW = NULL; } // not virtual
		bool IsOk()            { return (m_envStringsW != NULL); }
		LPWCH GetEnvStringsW() { return m_envStringsW; }
	};
	*/

	// On destruction, wipes the string with zeros. Used to clear passwords from memory.
	// Obviously, if the string is copied or re-allocated then bits of it may be left lying around.
	// Not much we can do without avoiding std::wstring entirely and re-writing a lot of stuff, which does not seem worth it.
#pragma optimize("", off)
	class StringWiper
	{
	private:
		std::wstring *m_pStr;
	private:
		StringWiper(const StringWiper &rhs); // disallow
		StringWiper &operator=(const StringWiper &rhs); // disallow
	public:
		StringWiper(std::wstring *pStr) : m_pStr(pStr) {}
		~StringWiper() { Wipe(); } // Not virtual.
		static void Wipe(std::wstring &str, wchar_t wipeChar = L'.')
		{
			str.assign(str.length(),wipeChar);
			str.clear(); 
		}
		void Wipe()
		{
			if (m_pStr)
			{
				Wipe(*m_pStr);
			}
		}
		void Forget() { m_pStr = NULL; }
	};
// optimization still off
	class StringWiperA
	{
	private:
		std::string *m_pStr;
	private:
		StringWiperA(const StringWiperA &rhs); // disallow
		StringWiperA &operator=(const StringWiperA &rhs); // disallow
	public:
		StringWiperA(std::string *pStr) : m_pStr(pStr) {}
		~StringWiperA() { Wipe(); } // Not virtual.
		static void Wipe(std::string &str, char wipeChar = '.')
		{
			str.assign(str.length(), wipeChar);
			str.clear(); 
		}
		void Wipe()
		{
			if (m_pStr)
			{
				Wipe(*m_pStr);
			}
		}
		void Forget() { m_pStr = NULL; }
	};
// optimization still off
	class BStrWiper
	{
	protected:
		BSTR m_bstr;
	private:
		BStrWiper(const BStrWiper &rhs); // disallow
		BStrWiper &operator=(const BStrWiper &rhs); // disallow
	public:
		BStrWiper(BSTR bstr) : m_bstr(bstr)           { }
		/* not virtual */ ~BStrWiper()                { Wipe(); Forget(); }
		      void Wipe()                             { ZeroMemory(m_bstr, ::SysStringLen(m_bstr) * sizeof(m_bstr[0])); } // SysStringLen returns zero if the bstr is null.
		      void Forget()                           { m_bstr = NULL; }
	};
// optimization still off
	class BufferWiper
	{
	protected:
		BYTE * m_b;
		size_t m_s;
	private:
		BufferWiper(const BufferWiper &rhs); // disallow
		BufferWiper &operator=(const BufferWiper &rhs); // disallow
	public:
		BufferWiper(BYTE *b, size_t s) : m_b(b), m_s(s) { }
		/* not virtual */ ~BufferWiper()                { Wipe(); Forget(); }
		      void Wipe()                               { ZeroMemory(m_b, m_s); } // SysStringLen returns zero if the bstr is null.
		      void Forget()                             { m_b = NULL; m_s = 0; }
	};
#pragma optimize("", on)

	class CoInitScoper
	{
	private:
		bool m_bSuccess{};
	public:
		CoInitScoper() { m_bSuccess = (::CoInitialize(NULL) == S_OK); }
		~CoInitScoper() { if (m_bSuccess) { ::CoUninitialize(); } }
		CoInitScoper(const CoInitScoper &) = delete;
		CoInitScoper &operator=(const CoInitScoper &) = delete;
		bool InitOK() { return m_bSuccess; }
	};

	class ScopedWaitCursor
	{
	private:
		bool m_fNeedRestore{};
	public:
		ScopedWaitCursor(bool fWaitCursorNow = true) { if (fWaitCursorNow) { SetWaitCursor(); } }
		~ScopedWaitCursor() { RestoreCursor(); }
		void SetWaitCursor() { ::SetCursor( ::LoadCursor(0,IDC_WAIT) ); m_fNeedRestore = true; }
		void RestoreCursor() { if (m_fNeedRestore) { m_fNeedRestore = false; ::SetCursor( ::LoadCursor(0,IDC_ARROW) ); } }

		ScopedWaitCursor(const ScopedWaitCursor &rhs) = delete;
		ScopedWaitCursor &operator=(const ScopedWaitCursor &rhs) = delete;
	};

	class LibraryScoper
	{
	private:
		HMODULE m_hLibrary{};
	public:
		LibraryScoper(const wchar_t *szLibraryPath, bool fAsDataFile)
			: m_hLibrary{ ::LoadLibraryEx(szLibraryPath, NULL, fAsDataFile ? LOAD_LIBRARY_AS_DATAFILE : 0) }
		{
		}
		LibraryScoper(const LibraryScoper &rhs) = delete;
		LibraryScoper &operator=(const LibraryScoper &rhs) = delete;
		~LibraryScoper()
		{
			if (m_hLibrary)
			{
				::FreeLibrary(m_hLibrary);
				m_hLibrary = NULL;
			}
		}
		bool InitOK()
		{
			return !!m_hLibrary;
		}
		HMODULE Get()
		{
			return m_hLibrary;
		}
		operator HMODULE()
		{
			return m_hLibrary;
		}
	};
};
